
/**
 * Позволяет хранить в ОЗУ не более {@link LRUCache#FILLED_SPACE_MAX_LENGTH}
 * элементов множества {@link LRUCache#list} и иметь к ним достаточно быстрый
 * доступ. Только при отсутствии в ОЗУ нужного элемента, происходит обращение к
 * {@link LRUCache#list}.
 *
 * @author quasiyoke
 */
public class LRUCache implements Collection{

private final int FILLED_SPACE_MAX_LENGTH;

private boolean isEmpty=true;

/**
 * Список, из которого можно брать элементы, которых нет в кеше.
 */
private Collection list;

/**
 * Смещение имеющегося подмножества внутри {@code objects}.
 */
private int filledSpaceArrayOffset=0;

/**
 * Смещение имеющегося подмножества в основном множестве.
 */
private int filledSpaceBigSetOffset=0;

/**
 * Хранилище подмножества.
 */
private Object[] objects;

/**
 * Создаёт {@code LRUCache}, количество элементов в котором при предельной
 * наполненности будет составлять {@code filledSpaceMaxLength}.
 * 
 * @throws IllegalArgumentException Если {@code filledSpaceMaxLength} не натуральное число.
 */
private LRUCache(int filledSpaceMaxLength){
    if(filledSpaceMaxLength<1){
        throw new IllegalArgumentException();
    }
    objects=new Object[FILLED_SPACE_MAX_LENGTH=filledSpaceMaxLength];
}

public LRUCache(int filledSpaceMaxLength,Collection list){
    this(filledSpaceMaxLength);
    this.list=list;
}

/**
 * @param setOffset Смещение в большом множестве.
 * 
 * @return Смещение внутри хранилища объектов {@link LRUCache#objects}.
 */
private int arrayOffset(int setOffset){
    int retval=(setOffset-filledSpaceBigSetOffset+filledSpaceArrayOffset)%
            FILLED_SPACE_MAX_LENGTH;
    return retval<0?retval+FILLED_SPACE_MAX_LENGTH:retval;
}

/**
 * @param index Номер элемента, который требуется получить. 
 * 
 * @return Элемент с заданным номером или <code>null</code>, если нет такового.
 */
public Object elementAt(int index) throws IndexOutOfSomeBoundsException{
    Object retval;
    if(index<filledSpaceBigSetOffset||index-filledSpaceBigSetOffset>=
            FILLED_SPACE_MAX_LENGTH||(retval=objects[arrayOffset(index)])==null){
        setElementAt(retval=list.elementAt(index),index);
    }
    return retval;
}

/**
 * Очищает кэш.
 */
public void removeAllElements(){
    if(!isEmpty){
        int j;
        for(j=0;j<FILLED_SPACE_MAX_LENGTH;j++){
            objects[j]=null;
        }
        isEmpty=true;
    }
}

/**
 * @param index Место объекта в большом множестве. Может быть любым целым числом.
 * 
 * @return Объект, находившийся ранее на этом месте, или {@code null} в случае
 * отсутствия такового.
 * 
 * @throws IndexOutOfSomeBoundsException Если для добавления нового элемента не
 * удаётся сохранить ни одного старого.
 * 
 * @see LRUCache#LRUCache(int)
 */
private Object setElementAt(Object object,int index) throws
        IndexOutOfSomeBoundsException{
    int j;
    /*
     * Если мы только добавили первый элемент.
     */
    if(isEmpty){
        objects[0]=object;
        filledSpaceBigSetOffset=index;
        filledSpaceArrayOffset=0;
        isEmpty=false;
        return null;
    /*
     * Если мы расширяем подмножество с начала.
     */
    }else{
        int indexArrayOffset=arrayOffset(index);
        if(index<filledSpaceBigSetOffset){
            /*
             * Не выпадает ли index мимо допустимых границ.
             */
            if(index>filledSpaceBigSetOffset-FILLED_SPACE_MAX_LENGTH){
                for(j=index;j<filledSpaceBigSetOffset;j++){
                    objects[arrayOffset(j)]=null;
                }
                objects[indexArrayOffset]=object;
                filledSpaceArrayOffset=indexArrayOffset;
                filledSpaceBigSetOffset=index;
                return null;
            }else{
                throw new IndexOutOfSomeBoundsException();
            }
        /*
         * Если мы расширяем подмножество с конца.
         */
        }else if(index>=filledSpaceBigSetOffset+FILLED_SPACE_MAX_LENGTH){
            /*
             * Не выпадает ли index мимо допустимых границ.
             */
            if(index<filledSpaceBigSetOffset+(FILLED_SPACE_MAX_LENGTH<<1)){
                if(indexArrayOffset<filledSpaceArrayOffset){
                    for(j=filledSpaceArrayOffset;j<FILLED_SPACE_MAX_LENGTH;j++){
                        objects[j]=null;
                    }
                    for(j=0;j<indexArrayOffset;j++){
                        objects[j]=null;
                    }
                }else{
                    for(j=filledSpaceArrayOffset;j<indexArrayOffset;j++){
                        objects[j]=null;
                    }
                }
                objects[indexArrayOffset]=object;
                filledSpaceArrayOffset=arrayOffset(index+1);
                filledSpaceBigSetOffset=index-FILLED_SPACE_MAX_LENGTH+1;
                return null;
            }else{
                throw new IndexOutOfSomeBoundsException();
            }
        /*
         * Если мы заменяем элемент в подмножестве.
         */
        }else{
            Object oldObject=objects[indexArrayOffset];
            objects[indexArrayOffset]=object;
            return oldObject;
        }
    }
}

/**
 * Меняет текущее главное хранилище объектов на заданное.
 */
public void setList(Collection list){
    if(!list.equals(this.list)){
        this.list=list;
        removeAllElements();
    }
}

}
